mod client;
mod messages;
mod prelude;
mod server;

use prelude::*;

struct Args {
    cmd: Subcommand,
}

enum Subcommand {
    Client(client::Args),
    Server(server::Args),
}

impl Args {
    fn new() -> Result<Self> {
        let mut args = std::env::args().skip(1);
        let cmd = match args.next().as_deref() {
            None => bail!("Need a subcommand, e.g. `client` or `server`"),
            Some("client") => {
                let mut client_args = client::Args::default();
                loop {
                    match args.next().as_deref() {
                        None => break,
                        Some("--period-ms") => {
                            client_args.period_ms =
                                u32::from_str(&args.next().context("Missing arg")?)?
                        }
                        Some("--port") => {
                            client_args.port = u16::from_str(&args.next().context("Missing arg")?)?
                        }
                        Some(_) => bail!("Invalid arg"),
                    }
                }
                Subcommand::Client(client_args)
            }
            Some("server") => {
                let mut server_args = server::Args::default();
                loop {
                    match args.next().as_deref() {
                        None => break,
                        Some("--port") => {
                            server_args.port = u16::from_str(&args.next().context("Missing arg")?)?
                        }
                        Some(_) => bail!("Invalid arg"),
                    }
                }
                Subcommand::Server(server_args)
            }
            Some(_) => bail!("Invalid subcommand"),
        };
        Ok(Self { cmd })
    }
}

#[tokio::main]
async fn main() -> Result<()> {
    let args = Args::new()?;
    tracing_subscriber::fmt::init();

    match args.cmd {
        Subcommand::Client(args) => run_client(args).await,
        Subcommand::Server(args) => run_server(args).await,
    }
}

async fn run_client(args: client::Args) -> Result<()> {
    let mut ctrl_c = signal(SignalKind::interrupt())?;
    let mut app = client::App::new(args).await?;
    loop {
        let ctl = poll_fn(|cx| {
            tracing::debug!("Loop");
            if ctrl_c.poll_recv(cx).is_ready() {
                tracing::debug!("Caught Ctrl-C");
                return Poll::Ready(ControlFlow::Break(Ok(())));
            }

            match app.poll_run(cx) {
                Poll::Pending => Poll::Pending,
                Poll::Ready(Ok(())) => Poll::Ready(ControlFlow::Continue(())),
                Poll::Ready(Err(err)) => Poll::Ready(ControlFlow::Break(Err(err))),
            }
        })
        .await;
        if let ControlFlow::Break(result) = ctl {
            return result;
        }
    }
}

async fn run_server(args: server::Args) -> Result<()> {
    let mut ctrl_c = signal(SignalKind::interrupt())?;
    let mut app = server::App::new(args).await?;
    poll_fn(|cx| {
        tracing::debug!("Loop");
        if ctrl_c.poll_recv(cx).is_ready() {
            tracing::debug!("Caught Ctrl-C");
            return Poll::Ready(Ok(()));
        }

        app.poll_run(cx)
    })
    .await
}
